from datetime import datetime
from typing import Sequence, Callable, Any
import json

from airflow.models import BaseOperator, SkipMixin
from airflow.providers.apache.kafka.hooks.consume import KafkaConsumerHook
from airflow.utils.context import Context
from airflow.providers.amazon.aws.hooks.s3 import S3Hook


class CustomConsumeFromTopicOperator(BaseOperator, SkipMixin):
    """

    """

    def __init__(
            self,
            topics: str | Sequence[str],
            kafka_config_id: str,
            apply_function_batch: Callable,
            max_messages: int = 100,
            poll_timeout: float = 60,
            batch_timeout: float = 86400,
            **kwargs,
    ):
        super().__init__(**kwargs)

        self.topics = topics
        self.kafka_config_id = kafka_config_id
        self.apply_function_batch = apply_function_batch
        self.max_messages = max_messages
        self.poll_timeout = poll_timeout
        self.batch_timeout = batch_timeout

    def execute(self, context: Context) -> Any:
        consumer = KafkaConsumerHook(
            topics=self.topics, kafka_config_id=self.kafka_config_id
        ).get_consumer()
        messages = []
        elapsed = 0

        while elapsed < self.batch_timeout:
            start_time = datetime.now()
            timeout = min(self.poll_timeout, self.batch_timeout - elapsed)

            messages = consumer.consume(
                num_messages=self.max_messages,
                timeout=timeout,
            )

            if len(messages) > 0:
                print(f"Received {len(messages)} messages")
                break

            elapsed += (datetime.now() - start_time).total_seconds()

        if not messages:
            self.log.info("No messages. Skipping downstream tasks...")
            if downstream_tasks := context['task'].get_flat_relatives(
                    upstream=False
            ):
                self.skip(
                    context['dag_run'],
                    context['ti'].execution_date,
                    downstream_tasks,
                )
            return

        result = self.apply_function_batch(messages, context)

        if result is not None:
            self.log.info(f"Result: {result}")
            context.get('ti').xcom_push(key='return_value', value=result)

        consumer.commit()
        consumer.close()
        return result


class ConsumeMessageDataOperator(CustomConsumeFromTopicOperator):
    MESSAGES_KEY = 'messages'
    MESSAGES_COUNT_KEY = 'messages_count'

    def __init__(self, **kwargs):
        args = {
            'apply_function_batch': self.messages_process,
        }
        args.update(kwargs)
        super().__init__(**args)

    @classmethod
    def messages_process(cls, messages: Sequence[Any], context) -> None:
        values = [message.value().decode('utf-8') for message in messages]
        context.get('ti').xcom_push(key=cls.MESSAGES_KEY, value=values)
        context.get('ti').xcom_push(key=cls.MESSAGES_COUNT_KEY, value=str(len(values)))


class ConsumeMessageDataXcomOperator(ConsumeMessageDataOperator):
    ...


class ConsumeMessageDataS3Operator(CustomConsumeFromTopicOperator):
    MESSAGES_KEY = 'messages'
    MESSAGES_COUNT_KEY = 'messages_count'
    S3_PATH_KEY = 's3_path'
    S3_BUCKET = 'airflow'

    def __init__(self, s3_config_id: str, **kwargs):
        args = {
            'apply_function_batch': self.messages_process,
        }
        args.update(kwargs)
        super().__init__(**args)

        self.s3_config_id = s3_config_id

    def messages_process(self, messages: Sequence[Any], context) -> None:
        s3_key = self._generate_s3_key(
            task_id=context.get('task').task_id,
            dag_id=context.get('dag').dag_id,
            run_id=context.get('dag_run').run_id,
            key=self.MESSAGES_KEY,
            map_index=0,
            suff="",
        )
        values = [json.loads(message.value().decode('utf-8')) for message in messages]

        context.get('ti').xcom_push(key=self.S3_PATH_KEY, value=s3_key)
        context.get('ti').xcom_push(key=self.MESSAGES_COUNT_KEY, value=str(len(values)))

        s3_hook = S3Hook(aws_conn_id=self.s3_config_id)
        s3_hook.load_string(
            string_data=json.dumps(values),
            bucket_name=self.S3_BUCKET,
            key=s3_key,
            replace=True,
        )

    @staticmethod
    def _generate_s3_key(task_id: str, dag_id: str, run_id: str, key: str,
                         map_index: int, suff: str, ) -> str:
        """Generates a unique S3 key for storing the data."""
        return f"xcom/{dag_id}/{run_id}/{task_id}/{key}_{map_index}.json{suff}"



import json
from datetime import datetime
from typing import Callable

from airflow.models import TaskInstance
from airflow.operators.python import PythonOperator
from airflow.providers.mysql.hooks.mysql import MySqlHook
from sqlalchemy import update, delete, and_
from sqlalchemy.inspection import inspect
from sqlalchemy.orm import sessionmaker, DeclarativeMeta, Session

from airflow_operators.utils.importer import import_from_string


class SQLAlchemySessionOperator(PythonOperator):
    def __init__(
            self, conn_id: str, python_callable: Callable, *args, **kwargs
    ):
        super().__init__(*args, python_callable=python_callable, **kwargs)
        self.conn_id = conn_id

    def get_session_factory(self) -> sessionmaker:
        hook = MySqlHook(self.conn_id)
        engine = hook.get_sqlalchemy_engine()

        return sessionmaker(bind=engine)

    def execute_callable(self):
        session_factory = self.get_session_factory()

        with session_factory() as session:
            try:
                result = self.python_callable(
                    *self.op_args, session=session, **self.op_kwargs
                )
            except Exception:
                session.rollback()
                raise
            else:
                session.commit()
        return result


class MergeOrmOperator(SQLAlchemySessionOperator):

    def __init__(
            self,
            conn_id: str,
            xcom_key: str,
            orm_model: DeclarativeMeta | str,
            *args,
            **kwargs,
    ):
        super().__init__(
            conn_id=conn_id,
            python_callable=self.save,
            *args,
            **kwargs,
        )

        self.xcom_key = xcom_key
        self.orm_model = (
            import_from_string(orm_model) if isinstance(orm_model, str) else orm_model
        )

    def save(self, session: Session, ti: TaskInstance):
        import json
        from airflow_operators.utils.orm_converter import OrmConverter

        org_jsons = ti.xcom_pull(key=self.xcom_key)

        for org_json in org_jsons:
            org_data = json.loads(org_json)
            organization = OrmConverter.convert_tree_to_orm(org_data, self.orm_model)
            session.merge(organization)


class OrmMergeOperator(MergeOrmOperator):
    ...


class UpdateOrmOperator(SQLAlchemySessionOperator):

    def __init__(
            self,
            conn_id: str,
            xcom_key: str,
            orm_model: DeclarativeMeta | str,
            condition_field: str,
            *args,
            **kwargs,
    ):
        super().__init__(
            conn_id=conn_id,
            python_callable=self.save,
            *args,
            **kwargs,
        )

        self.xcom_key = xcom_key
        self.orm_model = (
            import_from_string(orm_model) if isinstance(orm_model, str) else orm_model
        )
        self.condition_field = condition_field

    def save(self, session: Session, ti: TaskInstance):
        item_jsons = ti.xcom_pull(key=self.xcom_key)

        if not item_jsons:
            return

        mapper = inspect(self.orm_model)
        column_keys = [col.key for col in mapper.attrs]  # type: ignore

        for item_json in item_jsons:
            item = json.loads(item_json)
            condition_value = item.pop(self.condition_field)

            values = {
                key: item[key]
                for key in column_keys
                if key != self.condition_field and key in item
            }

            for upd_key in ("updated", "updated_at"):
                if upd_key in column_keys:
                    values[upd_key] = datetime.now()

            print(values)

            stmt = (
                update(self.orm_model)
                .where(getattr(self.orm_model, self.condition_field) == condition_value)
                .values(**values)
            )
            session.execute(stmt)


class DeleteOrmOperator(SQLAlchemySessionOperator):

    def __init__(
            self,
            conn_id: str,
            xcom_key: str,
            orm_model: DeclarativeMeta | str,
            *args,
            **kwargs,
    ):
        super().__init__(
            conn_id=conn_id,
            python_callable=self.save,
            *args,
            **kwargs,
        )

        self.xcom_key = xcom_key
        self.orm_model = (
            import_from_string(orm_model) if isinstance(orm_model, str) else orm_model
        )

    def save(self, session: Session, ti: TaskInstance):
        item_jsons = ti.xcom_pull(key=self.xcom_key)

        if not item_jsons:
            return

        mapper = inspect(self.orm_model)
        pk_columns = mapper.primary_key  # type: ignore

        if not pk_columns:
            raise ValueError("Модель не имеет первичного ключа")

        for item_json in item_jsons:
            item_data = json.loads(item_json)

            conditions = [col == item_data[col.key] for col in pk_columns]
            stmt = delete(self.orm_model).where(and_(*conditions))
            session.execute(stmt)
